﻿//-------------------------------------------------------------------------- 
//
//  Copyright (c) Microsoft Corporation.  All rights reserved. 
//
//  File: SqlExpressClientSyncProvider.cs 
//
//  Description: CUSTOMIZED Generic client synchronization provider.
//
//--------------------------------------------------------------------------
using Microsoft.Synchronization.Data;
using Microsoft.Synchronization.Data.Server;
using System;
using System.Collections.Generic;
using System.Data;
using System.Data.Sql;
using System.Data.SqlClient;
using System.IO;
using System.Text;
using System.Data.Common;

namespace Microsoft.Synchronization.Data.SqlExpress
{
    ///<summary>
    /// A CUSTOMIZED generic client sync provider that can connect to SQL Express
    /// </summary>
    /// <remarks>    
    /// SqlExpressClientSyncProvider inherits from ClientSyncProvider, and it is a generic 
    /// implementation of ServerSyncProvider. SqlExpressClientSyncProvider uses the mechanisms
    /// of DbServerSyncProvider to connect to the client (ie using a DBConnection)
    /// </remarks>
    public class SqlExpressClientSyncProvider : ClientSyncProvider
    {

        #region Constructor

        /// <summary>
        /// Default constructor
        /// </summary>
        public SqlExpressClientSyncProvider()
        {
            _dbSyncProvider = new DbServerSyncProvider();
            _dbSyncProvider.ApplyingChanges += new EventHandler<ApplyingChangesEventArgs>(OnApplyingChanges);
            _dbSyncProvider.ApplyChangeFailed += new EventHandler<ApplyChangeFailedEventArgs>(OnApplyChangeFailed);
            _clientId = Guid.Empty;
            _refCntSession = 0;
            _transaction = null;
        }

        #endregion

        #region Statics & Constants

        public const string TOKEN_TABLE_NAME = "|%TableName%|";
        public const string PARAMETER_SERVER_ID = "__server_id";
        public const string PARAMETER_CLIENT_ID = "sync_client_id";
        public const string PARAMETER_TABLE_RECEIVED_ANCHOR = "sync_table_received_anchor";
        public const string PARAMETER_TABLE_SENT_ANCHOR = "sync_table_sent_anchor";
        public const string PARAMETER_CURRENT_TABLE = "sync_current_table";
        public const string PARAMETER_ANCHOR_VALUE = "sync_anchor";
        public const string PARAMETER_ANCHOR_SERVER_APPLIED = "sync_anchor_server_applied";
        public const string PARAMETER_ACTUAL_CLIENT_ID = "actual_binary_client_id";

        // duplicate of the function found in syncutil
        public static object GetSyncObjectOutParameter(string parameter, IDbCommand command)
        {
            bool flag;
            return GetSyncObjectOutParameter(parameter, command, out flag);
        }
        // duplicate of the function found in syncutil
        public static object GetSyncObjectOutParameter(string parameter, IDbCommand command, out bool found)
        {
            found = true;
            DbParameter parameter2 = GetParameter(command, parameter);
            if (parameter2 != null)
            {
                return parameter2.Value;
            }
            found = false;
            return null;
        }
        // duplicate of the function found in syncutil
        public static DbParameter GetParameter(IDbCommand command, string parameterName)
        {
            if (command != null)
            {
                if (command.Parameters.Contains("@" + parameterName))
                {
                    return (DbParameter)command.Parameters["@" + parameterName];
                }
                if (command.Parameters.Contains(":" + parameterName))
                {
                    return (DbParameter)command.Parameters[":" + parameterName];
                }
                if (command.Parameters.Contains(parameterName))
                {
                    return (DbParameter)command.Parameters[parameterName];
                }
            }
            return null;
        }

        #endregion

        #region Events & Invokers

        public event EventHandler<ApplyChangeFailedEventArgs> ApplyChangeFailed;
        public event EventHandler<ApplyingChangesEventArgs> ApplyingChanges;

        protected virtual void OnApplyChangeFailed(object sender, ApplyChangeFailedEventArgs e)
        {
            if (ApplyChangeFailed != null)
            {
                ApplyChangeFailed(sender, e);
            }
        }
        protected virtual void OnApplyingChanges(object sender, ApplyingChangesEventArgs e)
        {
            // Roll ApplyChanges() modifications done by inner _dbSyncProvider into one transaction with anchor changes
            if (_transaction != null)
            {
                e.Transaction = _transaction;
            }

            if (ApplyingChanges != null)
            {
                ApplyingChanges(sender, e);
            }
        }

        #endregion

        #region Private Properties

        private DbServerSyncProvider _dbSyncProvider;
        private Guid _clientId;

        private int _refCntSession;
        private IDbTransaction _transaction;

        private IDbCommand _selectClientIDCommand;
        private IDbCommand _selectTableReceivedAnchorCommand;
        private IDbCommand _selectTableSentAnchorCommand;
        private IDbCommand _updateTableSentAnchorCommand;
        private IDbCommand _updateTableReceivedAnchorCommand;
        private IDbCommand _updateTableServerAppliedAnchorCommand;

        private Dictionary<string, SyncAnchor> _downloadedTables;

        private bool _isSqlChangeTracking;

        #endregion

        #region Public Properties

        /// <summary>
        /// Sets whether this instance is using Sql Change Tracking. (if true, UpdateActualSentAnchorValues is called after the transaction is committed)
        /// </summary>
        public bool SqlChangeTracking
        {
            get
            {
                return _isSqlChangeTracking;
            }
            set
            {
                _isSqlChangeTracking = value;
            }
        }

        /// <summary>
        /// Gets or sets the command that will return the new anchor value. 
        /// </summary>
        /// <value>
        /// A <b>IDbCommand</b>-inheritated object.
        /// </value>
        public IDbCommand SelectNewAnchorCommand
        {
            get
            {
                return _dbSyncProvider.SelectNewAnchorCommand;
            }
            set
            {
                _dbSyncProvider.SelectNewAnchorCommand = value;
            }
        }

        /// <summary>
        /// Gets or sets the server connection object.
        /// </summary>
        /// <value>
        /// A <b>DbConnection</b> object.
        /// </value>
        public IDbConnection Connection
        {
            get
            {
                return _dbSyncProvider.Connection;
            }
            set
            {
                if (value == null)
                {
                    throw new ArgumentNullException("value");
                }

                if (value.State == ConnectionState.Closed)
                {
                    // giving the connection to the _dbSyncPRovider as open 
                    // will prevent it from closing off the connection at end 
                    // of operations
                    value.Open();
                    _dbSyncProvider.Connection = value;
                    value.Close();
                }
                else _dbSyncProvider.Connection = value;

            }
        }

        /// <summary>
        /// Gets the collection of <b>SyncAdapter</b>.
        /// </summary>
        /// <value>
        /// A <b>SyncAdapterCollection</b> object.
        /// </value>
        public SyncAdapterCollection SyncAdapters
        {
            get
            {
                return _dbSyncProvider.SyncAdapters;
            }
        }

        /// <summary>
        /// Gets and sets ClientId
        /// </summary>
        /// <remarks>
        /// A new client is generated and saved if none is present
        /// </remarks>
        /// <value>
        /// A <b>Guid</b> object.
        /// </value>
        public override Guid ClientId
        {
            get
            {
                this.EnsureClientID();
                return _clientId;
            }
            set
            {
                _clientId = value;
            }
        }

        /// <summary>
        /// Gets or sets the command that will return the current client ID (GUID). 
        /// </summary>
        /// <value>
        /// An <b>IDbCommand</b>-inheritated object.
        /// </value>
        /// <remarks>Expects a scalar value or a parametized command using SqlExpressClientSyncProvider.PARAMETER_CLIENT_ID</remarks>
        public IDbCommand SelectClientIDCommand
        {
            get
            {
                return _selectClientIDCommand;
            }
            set
            {
                _selectClientIDCommand = value;
            }
        }

        /// <summary>
        /// Gets or sets the command that will return the current table received anchor. 
        /// </summary>
        /// <value>
        /// An <b>IDbCommand</b>-inheritated object.
        /// </value>
        /// <remarks>Expects a scalar value or a parametized command using SqlExpressClientSyncProvider.PARAMETER_TABLE_RECEIVED_ANCHOR</remarks>
        public IDbCommand SelectTableReceivedAnchorCommand
        {
            get
            {
                return _selectTableReceivedAnchorCommand;
            }
            set
            {
                _selectTableReceivedAnchorCommand = value;
            }
        }
        /// <summary>
        /// Gets or sets the command that will return the current table sent anchor. 
        /// </summary>
        /// <value>
        /// An <b>IDbCommand</b>-inheritated object.
        /// </value>
        /// <remarks>Expects a scalar value or a parametized command using SqlExpressClientSyncProvider.PARAMETER_TABLE_SENT_ANCHOR</remarks>
        public IDbCommand SelectTableSentAnchorCommand
        {
            get
            {
                return _selectTableSentAnchorCommand;
            }
            set
            {
                _selectTableSentAnchorCommand = value;
            }
        }

        /// <summary>
        /// Gets or sets the command that will update the current table sent anchor. 
        /// </summary>
        /// <value>
        /// An <b>IDbCommand</b>-inheritated object.
        /// </value>
        /// <remarks>Expects a parametized command using SqlExpressClientSyncProvider.PARAMETER_CURRENT_TABLE and SqlExpressClientSyncProvider.PARAMETER_ANCHOR_VALUE</remarks>
        public IDbCommand UpdateTableSentAnchorCommand
        {
            get
            {
                return _updateTableSentAnchorCommand;
            }
            set
            {
                _updateTableSentAnchorCommand = value;
            }
        }
        /// <summary>
        /// Gets or sets the command that will update the current table received anchor. 
        /// </summary>
        /// <value>
        /// An <b>IDbCommand</b>-inheritated object.
        /// </value>
        /// <remarks>Expects a parametized command using SqlExpressClientSyncProvider.PARAMETER_CURRENT_TABLE and SqlExpressClientSyncProvider.PARAMETER_ANCHOR_VALUE</remarks>
        public IDbCommand UpdateTableReceivedAnchorCommand
        {
            get
            {
                return _updateTableReceivedAnchorCommand;
            }
            set
            {
                _updateTableReceivedAnchorCommand = value;
            }
        }

        /// <summary>
        /// Gets or sets the command that will update the current table server applied anchor. [for race conditions during sync process]
        /// </summary>
        /// <value>
        /// An <b>IDbCommand</b>-inheritated object.
        /// </value>
        /// <remarks>Expects a parametized command using SqlExpressClientSyncProvider.PARAMETER_CURRENT_TABLE and SqlExpressClientSyncProvider.PARAMETER_ANCHOR_SERVER_APPLIED</remarks>
        public IDbCommand UpdateTableServerAppliedAnchorCommand
        {
            get
            {
                return _updateTableServerAppliedAnchorCommand;
            }
            set
            {
                _updateTableServerAppliedAnchorCommand = value;
            }
        }


        #endregion

        #region Public Overrides

        /// <summary>
        /// Begin a transaction. This method is invoked to mark atomic operations.   
        /// </summary>
        /// <param name="syncSession">SyncSession information</param>               
        public override void BeginTransaction(SyncSession syncSession)
        {
            _refCntSession++;
            if (Connection.State == ConnectionState.Closed)
            {
                Connection.Open();
            }
            if (_transaction == null)
            {
                _transaction = _dbSyncProvider.Connection.BeginTransaction();
            }
        }

        /// <summary>
        /// End transaction. This method is called by the agent at to conclude an atomic operation.
        /// </summary>
        /// <param name="commit">Commit/Abort SyncSession</param>               
        /// <param name="syncSession">SyncSession information</param>               
        public override void EndTransaction(bool commit, SyncSession syncSession)
        {
            _refCntSession--;
            if (_refCntSession == 0)
            {
                if (commit)
                    _transaction.Commit();
                else
                    _transaction.Rollback();

                _transaction.Dispose();
                _transaction = null;

                Connection.Close();
            }
            if (_refCntSession < 0) _refCntSession = 0;

            // Added to fix all actual sent anchors.. Outside of the sync transaction, but thats a flaw we'll have to live with.
            if (SqlChangeTracking && (_refCntSession == 0) && commit) // If last transaction
            {
                // If we have tables to clean up
                if ((_downloadedTables != null))
                {
                    Dictionary<string, SyncAnchor> tablesUpdated = _downloadedTables;
                    _downloadedTables = null;
                    UpdateServerAppliedAnchorValues(tablesUpdated, GetServerID(syncSession));
                }

            }
            //else if (_refCntSession < 0)
            //    throw new SyncException("Transaction error: _refCntSession < 0");
        }


        /// <summary>
        /// Creates the database schema on client database -- NOT IMPLEMENTED
        /// </summary>
        /// <remarks>
        /// In the current implementation of this class, we assume that the 
        /// client already has the same schema as the server (run the demo scripts).
        /// </remarks>
        public override void CreateSchema(SyncTable syncTable, SyncSchema syncSchema)
        {
            throw new NotSupportedException("Create Schema is not supported in this version."
                    + "Please make sure client and server have same schema!");
        }

        /// <summary>
        /// Gets the changes made on the client since last sync.
        /// </summary>
        /// <param name="groupMetadata"> Contains table metadata </param>
        /// <param name="syncSession"> The current sync session </param>
        /// <returns> SyncContext populated with the incremental changes </returns>
        public override SyncContext GetChanges(SyncGroupMetadata groupMetadata, SyncSession syncSession)
        {

            // neet to set the LastReceivedAnchor as the LastSentAnchor since 
            // DbServerSyncProvider operates from the server's perspective, so
            // we swap the two fields temporarily. 
            foreach (SyncTableMetadata metaTable in groupMetadata.TablesMetadata)
            {
                SyncAnchor temp = metaTable.LastReceivedAnchor;
                metaTable.LastReceivedAnchor = metaTable.LastSentAnchor;
                metaTable.LastSentAnchor = temp;
            }

            SyncContext context = _dbSyncProvider.GetChanges(groupMetadata, syncSession);

            //swap them back for consistency
            foreach (SyncTableMetadata metaTable in groupMetadata.TablesMetadata)
            {
                SyncAnchor temp = metaTable.LastReceivedAnchor;
                metaTable.LastReceivedAnchor = metaTable.LastSentAnchor;
                metaTable.LastSentAnchor = temp;
            }
            return context;
        }


        /// <summary>
        /// Apply changes downloaded from the server. 
        /// </summary>
        /// <remarks>
        /// Inner _dbSyncProvider will take care of applying changes to actual 
        /// data, but we need to take care of updating anchor metadata. 
        /// </remarks>
        /// <param name="groupMetadata"> Contains table metadata info </param>
        /// <param name="dataSet"> Contains changes to be applied </param>
        /// <param name="syncSession"> Current sync session </param>
        /// <returns> SyncContext object to Sync Agent </returns>
        public override SyncContext ApplyChanges(SyncGroupMetadata groupMetadata, DataSet dataSet, SyncSession syncSession)
        {
            EnsureServerIDParameter(syncSession);

            //Map SyncDirection from client POV to our internal server POV

            foreach (SyncTableMetadata tableMetadata in groupMetadata.TablesMetadata)
            {
                if (tableMetadata.SyncDirection == SyncDirection.DownloadOnly || tableMetadata.SyncDirection == SyncDirection.Snapshot)
                {
                    //This SyncDirection DownloadOnly/Snapshot is from a Client point of view. But our client is inturn a Server provider.   Hence switch this to UploadOnly
                    tableMetadata.SyncDirection = SyncDirection.UploadOnly;
                }

                else if (tableMetadata.SyncDirection == SyncDirection.UploadOnly)
                {
                    //This SyncDirection UploadOnly is from Client POV. But our client is inturn a Server provider. Hence switch this to DownloadOnly
                    tableMetadata.SyncDirection = SyncDirection.DownloadOnly;
                }
            }

            // neet to set the LastReceivedAnchor as the LastSentAnchor since 
            // DbServerSyncProvider operates from the server's perspective, so
            // we swap the two fields temporarily. 
            // Note that even if we do this, the NewAnchor value will be the one
            // from the server, not local which is invalid since the client and server
            // clocks are always at least the tiniest bit misaligned
            foreach (SyncTableMetadata metaTable in groupMetadata.TablesMetadata)
            {
                SyncAnchor temp = metaTable.LastReceivedAnchor;
                metaTable.LastReceivedAnchor = metaTable.LastSentAnchor;
                metaTable.LastSentAnchor = temp;
            }

            // SyncTracer.Verbose("New Anchor: " + DeserializeAnchorValue(groupMetadata.NewAnchor.Anchor));

            // Swap ClientID with ServerID Value
            Guid actualClientID = syncSession.ClientId;
            Guid actualServerID = GetServerID(syncSession);
            syncSession.ClientId = actualServerID;
            SyncContext syncContext = _dbSyncProvider.ApplyChanges(groupMetadata, dataSet, syncSession);
            // Swap ServerID with ClientID
            syncSession.ClientId = actualClientID;

            // SyncTracer.Verbose("New Anchor: " + DeserializeAnchorValue(groupMetadata.NewAnchor.Anchor));

            //swap them back for consistency
            foreach (SyncTableMetadata metaTable in groupMetadata.TablesMetadata)
            {
                SyncAnchor temp = metaTable.LastReceivedAnchor;
                metaTable.LastReceivedAnchor = metaTable.LastSentAnchor;
                metaTable.LastSentAnchor = temp;
            }
            if (SqlChangeTracking)
            {
                _downloadedTables = new Dictionary<string, SyncAnchor>(StringComparer.InvariantCultureIgnoreCase);
            }
            foreach (SyncTableMetadata table in groupMetadata.TablesMetadata)
            {
                SetTableReceivedAnchor(table.TableName, groupMetadata.NewAnchor);

                if (SqlChangeTracking)
                {
                    // Server updated these tables, so we need to update sent (so we don't echo it back later)
                    _downloadedTables[table.TableName] = table.LastSentAnchor;

                }
            }

            return syncContext;

        }


        /// <summary>
        /// Disposes this Client Sync Provider instance 
        /// </summary>
        public override void Dispose()
        {
            this.Dispose(true);
            GC.SuppressFinalize(this);
        }

        #endregion

        #region Protected Methods

        /// <summary>
        /// Ensures any anchor commands that use ServerID are updated
        /// </summary>
        protected virtual void EnsureServerIDParameter(SyncSession syncSession)
        {
            byte[] serverID = SerializeAnchorValue(GetServerID(syncSession)); // this may be used by masked _dbSyncProvider

            // Anchor Commands
            DbParameter serverIdParam = GetParameter(UpdateTableReceivedAnchorCommand, PARAMETER_SERVER_ID);
            if (serverIdParam != null)
            {
                serverIdParam.Value = serverID;
            }
            serverIdParam = GetParameter(UpdateTableSentAnchorCommand, PARAMETER_SERVER_ID);
            if (serverIdParam != null)
            {
                serverIdParam.Value = serverID;
            }
        }

        /// <summary>
        /// Gets the current serverID or generates one if not given.
        /// </summary>
        protected virtual Guid GetServerID(SyncSession syncSession)
        {
            if (!syncSession.SyncParameters.Contains(PARAMETER_SERVER_ID))
            {
                syncSession.SyncParameters.Add(PARAMETER_SERVER_ID, Guid.NewGuid()); // if not there: create random guid for the lifetime of the session
            }
            return (Guid)syncSession.SyncParameters[PARAMETER_SERVER_ID].Value;
        }

        /// <summary>
        /// Generates the command needed to retrieve the lateset server applied anchor values
        /// </summary>
        protected virtual IDbCommand GenerateServerAppliedAnchorAnchorValueCommand(string tableName)
        {
            // create command
            string parameterAnchorValue = string.Format("@{0}", PARAMETER_ANCHOR_VALUE);
            string parameterServerID = string.Format("@{0}", PARAMETER_SERVER_ID);

            string parameterAnchorServerApplied = string.Format("@{0}", PARAMETER_ANCHOR_SERVER_APPLIED);

            string commandText = string.Format("SELECT {3} = ISNULL(MAX(SYS_CHANGE_CREATION_VERSION), {0}) "
                + " FROM CHANGETABLE(CHANGES {1}, {0}) CT "
                + " WHERE CT.SYS_CHANGE_CONTEXT = {2} ", parameterAnchorValue, tableName, parameterServerID, parameterAnchorServerApplied);

            SqlCommand sqlCommand = new SqlCommand(commandText);
            sqlCommand.Parameters.Add(parameterAnchorValue, SqlDbType.BigInt);
            sqlCommand.Parameters.Add(parameterAnchorServerApplied, SqlDbType.BigInt);
            sqlCommand.Parameters[parameterAnchorServerApplied].Direction = ParameterDirection.Output;
            sqlCommand.Parameters.Add(parameterServerID, SqlDbType.VarBinary, 16);

            return sqlCommand;
        }

        /// <summary>
        /// Retrieves the actual sent anchor after applying downloads from the server
        /// </summary>
        protected virtual SyncAnchor GetServerAppliedAnchorValue(string tableName, Guid serverID, SyncAnchor tableAnchor)
        {
            IDbCommand command = GenerateServerAppliedAnchorAnchorValueCommand(tableName);
            SyncAnchor serverAppliedAnchor = null;

            object serverAppliedAnchorValue = null;

            bool commandPassed = false;

            try
            {
                BeginTransaction(null);
                command.Connection = _dbSyncProvider.Connection;
                command.Transaction = _transaction;

                DbParameter anchorParam = GetParameter(command, PARAMETER_ANCHOR_VALUE);
                if (anchorParam != null)
                {
                    anchorParam.Value = DeserializeAnchorValue(tableAnchor.Anchor);
                }
                DbParameter serverParam = GetParameter(command, PARAMETER_SERVER_ID);
                if (serverParam != null)
                {
                    serverParam.Value = serverID.ToByteArray();
                }

                if (command.Connection.State == ConnectionState.Closed)
                {
                    command.Connection.Open();
                }

                // parameter mode
                command.ExecuteNonQuery();
                serverAppliedAnchorValue = GetSyncObjectOutParameter(PARAMETER_ANCHOR_SERVER_APPLIED, command);
                commandPassed = true;
            }
            finally
            {
                EndTransaction(commandPassed, null);
            }

            if ((serverAppliedAnchorValue == null) || (serverAppliedAnchorValue == System.DBNull.Value))
            {
                serverAppliedAnchor = tableAnchor; // inputted value
            }
            else
            {
                serverAppliedAnchor = new SyncAnchor(SerializeAnchorValue(serverAppliedAnchorValue));
            }

            return serverAppliedAnchor;
        }

        /// <summary>
        /// Processes all downloaded tables sent parameters
        /// </summary>
        protected virtual void UpdateServerAppliedAnchorValues(Dictionary<string, SyncAnchor> tablesToUpdate, Guid serverID)
        {
            foreach (string tableName in tablesToUpdate.Keys)
            {
                SyncAnchor serverAppliedAnchor = GetServerAppliedAnchorValue(tableName, serverID, tablesToUpdate[tableName]);

                // if we downloaded changes [max server anchor != last sent anchor]
                if (DeserializeAnchorValue(serverAppliedAnchor.Anchor).Equals(DeserializeAnchorValue(tablesToUpdate[tableName].Anchor)))
                {
                    // we didnt download anything, so nothing to log ( or something else got us here... ?)
                    serverAppliedAnchor = null;
                }
                SetTableServerAppliedAnchor(tableName, serverAppliedAnchor);
            }
        }

        /// <summary>
        /// If no ClientID is provided, executes SelectClientIDCommand
        /// </summary>
        protected virtual void EnsureClientID()
        {
            if (_clientId == Guid.Empty)
            {
                if (SelectClientIDCommand == null)
                {
                    throw new NotImplementedException("Unable to acquire ClientID. You must either provide it via the ClientID property or by providing a SelectClientIDCommand");
                }

                bool commandPassed = false;

                string paramName = string.Format("@{0}", PARAMETER_CLIENT_ID);

                IDataReader reader = null;
                try
                {
                    BeginTransaction(null);
                    SelectClientIDCommand.Connection = _dbSyncProvider.Connection;
                    SelectClientIDCommand.Transaction = _transaction;

                    if (SelectClientIDCommand.Connection.State == ConnectionState.Closed)
                    {
                        SelectClientIDCommand.Connection.Open();
                    }
                    if (GetParameter(SelectClientIDCommand, PARAMETER_CLIENT_ID) != null)
                    {
                        // parameter mode
                        SelectClientIDCommand.ExecuteNonQuery();
                        object result = GetSyncObjectOutParameter(PARAMETER_CLIENT_ID, SelectClientIDCommand);

                        if ((result != null) && (result != DBNull.Value))
                        {
                            _clientId = new Guid(result.ToString());
                        }
                    }
                    else
                    {
                        // assume scalar mode
                        reader = SelectClientIDCommand.ExecuteReader();
                        if (reader.Read())
                        {
                            _clientId = reader.GetGuid(0);
                            reader.Close();
                        }
                    }
                    commandPassed = true;
                }
                finally
                {
                    try
                    {
                        if (reader != null)
                        {
                            if (!reader.IsClosed)
                            {
                                reader.Close();
                            }
                            reader.Dispose();
                        }
                    }
                    finally
                    {
                        EndTransaction(commandPassed, null);
                    }
                }
            }
        }

        /// <summary>
        /// Disposes this Client Sync Provider instance 
        /// </summary>
        /// <param name="disposing">True if explicit finalization, false if through GC</param>
        //[SuppressMessage("Microsoft.Reliability", "CA2004:RemoveCallsToGCKeepAlive")]
        protected virtual void Dispose(bool disposing)
        {
            _dbSyncProvider.Dispose();
        }

        protected virtual byte[] SerializeAnchorValue(object anchorVal)
        {
            using (MemoryStream serializationStream = new MemoryStream())
            {
                new System.Runtime.Serialization.Formatters.Binary.BinaryFormatter().Serialize(serializationStream, anchorVal);
                return serializationStream.ToArray();
            }
        }
        protected virtual object DeserializeAnchorValue(byte[] anchor)
        {
            using (MemoryStream serializationStream = new MemoryStream(anchor))
            {
                return new System.Runtime.Serialization.Formatters.Binary.BinaryFormatter().Deserialize(serializationStream);
            }
        }

        protected virtual SyncAnchor ExecuteAnchorSelectCommand(IDbCommand command, string outParameterName, string tableName)
        {
            object anchorVal = null;
            bool commandPassed = false;

            try
            {
                BeginTransaction(null);
                command.Connection = _dbSyncProvider.Connection;
                command.Transaction = _transaction;

                DbParameter tableParam = GetParameter(command, PARAMETER_CURRENT_TABLE);
                if (tableParam != null)
                {
                    tableParam.Value = tableName;
                }
                if (command.Connection.State == ConnectionState.Closed)
                {
                    command.Connection.Open();
                }
                if (GetParameter(command, outParameterName) != null)
                {
                    // parameter mode
                    command.ExecuteNonQuery();
                    anchorVal = GetSyncObjectOutParameter(outParameterName, command);
                }
                else
                {
                    // assume scalar mode
                    anchorVal = command.ExecuteScalar();
                }
                commandPassed = true;
            }
            finally
            {
                EndTransaction(commandPassed, null);
            }

            if ((anchorVal == null) || (anchorVal == System.DBNull.Value))
            {
                return new SyncAnchor();
            }
            else
            {
                return new SyncAnchor(SerializeAnchorValue(anchorVal));
            }
        }
        protected virtual void ExecuteAnchorUpdateCommand(IDbCommand command, string tableName, SyncAnchor anchor)
        {
            bool commandPassed = false;
            try
            {
                BeginTransaction(null);
                command.Connection = _dbSyncProvider.Connection;
                command.Transaction = _transaction;

                DbParameter tableParam = GetParameter(command, PARAMETER_CURRENT_TABLE);
                if (tableParam != null)
                {
                    tableParam.Value = tableName;
                }
                DbParameter anchorParam = GetParameter(command, PARAMETER_ANCHOR_VALUE);
                if (anchorParam != null)
                {
                    anchorParam.Value = DeserializeAnchorValue(anchor.Anchor);
                }

                if (command.Connection.State == ConnectionState.Closed)
                {
                    command.Connection.Open();
                }
                if (command.ExecuteNonQuery() == 0)
                {
                    throw new Exception("Unable to complete sync. SetTableReceivedAnchor() had no effect. Verify anchor table is aware of table: " + tableName);
                }
                commandPassed = true;
            }
            finally
            {
                EndTransaction(commandPassed, null);
            }

        }


        #endregion

        #region Public Overrides

        /// <summary>
        /// Retrieves the last received anchor from the 'anchor' metatable.
        /// </summary>
        /// <param name="tableName"> The name of the table which we want the anchor for. </param>
        /// <returns> A sync anchor object containing the last received anchor. </returns>
        public override SyncAnchor GetTableReceivedAnchor(string tableName)
        {
            if (SelectTableReceivedAnchorCommand == null)
            {
                throw new NotImplementedException("You must provide a SelectTableReceivedAnchorCommand");
            }

            return ExecuteAnchorSelectCommand(SelectTableReceivedAnchorCommand, PARAMETER_TABLE_RECEIVED_ANCHOR, tableName);
        }
        /// <summary>
        /// Retrieves the last sent anchor from the 'anchor' metatable.
        /// </summary>
        /// <param name="tableName"> The name of the table for which we want the anchor. </param>
        /// <returns> A sync anchor object containing the last sent anchor. </returns>
        public override SyncAnchor GetTableSentAnchor(string tableName)
        {
            if (SelectTableSentAnchorCommand == null)
            {
                throw new NotImplementedException("You must provide a SelectTableSentAnchorCommand");
            }
            return ExecuteAnchorSelectCommand(SelectTableSentAnchorCommand, PARAMETER_TABLE_SENT_ANCHOR, tableName);
        }

        /// <summary>
        /// Sets the last received anchor in the 'anchor' metatable
        /// </summary>
        /// <param name="tableName"> The name of the table for which we want to set the anchor </param>
        /// <param name="anchor"> SyncAnchor object containing the anchor. </param>
        public override void SetTableReceivedAnchor(string tableName, SyncAnchor anchor)
        {
            if (UpdateTableReceivedAnchorCommand == null)
            {
                throw new NotImplementedException("You must provide a UpdateTableReceivedAnchorCommand");
            }
            ExecuteAnchorUpdateCommand(UpdateTableReceivedAnchorCommand, tableName, anchor);
        }

        /// <summary>
        /// Sets the last sent anchor in the 'anchor' metatable
        /// </summary>
        /// <param name="tableName"> The name of the table for whcih we want to set the anchor </param>
        /// <param name="anchor"> SyncAnchor object containing the anchor. </param>
        public override void SetTableSentAnchor(string tableName, SyncAnchor anchor)
        {
            if (UpdateTableSentAnchorCommand == null)
            {
                throw new NotImplementedException("You must provide a UpdateTableSentAnchorCommand");
            }


            ExecuteAnchorUpdateCommand(UpdateTableSentAnchorCommand, tableName, anchor);
        }
        /// <summary>
        /// Sets the last server applied anchor in the 'anchor' metatable
        /// </summary>
        /// <param name="tableName">The name of the table for which we want to set the anchor </param>
        /// <param name="serverAnchor"> SyncAnchor object containing the anchor last applied by the server. </param>
        public virtual void SetTableServerAppliedAnchor(string tableName, SyncAnchor serverAppliedAnchor)
        {
            if (UpdateTableServerAppliedAnchorCommand == null)
            {
                throw new NotImplementedException("You must provide a UpdateTableServerAppliedAnchorCommand");
            }

            bool commandPassed = false;
            IDbCommand command = UpdateTableServerAppliedAnchorCommand;
            try
            {
                BeginTransaction(null);
                command.Connection = _dbSyncProvider.Connection;
                command.Transaction = _transaction;

                DbParameter tableParam = GetParameter(command, PARAMETER_CURRENT_TABLE);
                if (tableParam != null)
                {
                    tableParam.Value = tableName;
                }
                DbParameter anchorAppliedParam = GetParameter(command, PARAMETER_ANCHOR_SERVER_APPLIED);
                if (anchorAppliedParam != null)
                {
                    if (serverAppliedAnchor == null)
                    {
                        anchorAppliedParam.Value = DBNull.Value;
                    }
                    else
                    {
                        anchorAppliedParam.Value = DeserializeAnchorValue(serverAppliedAnchor.Anchor);
                    }
                }

                if (command.Connection.State == ConnectionState.Closed)
                {
                    command.Connection.Open();
                }
                if (command.ExecuteNonQuery() == 0)
                {
                    throw new Exception("Unable to complete sync. SetTableServerAppliedAnchor() had no effect. Verify anchor table is aware of table: " + tableName);
                }
                commandPassed = true;
            }
            finally
            {
                EndTransaction(commandPassed, null);
            }
        }
        
        #endregion


    }
}
