﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Synchronization;
using Microsoft.Synchronization.Data;
using System.Windows.Forms;
using SQLChangeTrackingSync.Sample.Entities;
using SQLChangeTrackingSync.Sample.Providers;

namespace SQLChangeTrackingSync.Sample
{
    public class SampleSyncAgent : SyncAgent
    {
        #region Constructor

        public SampleSyncAgent()
        {
        }

        #endregion

        #region Private Properties

        private bool _initialized;
        private Guid _requestingClientID;
        private Dictionary<string, ApplyAction> _syncPolicy;
        private List<ConflictLog> _conflicts = new List<ConflictLog>();

        #endregion

        #region Public Properties
        
        public Guid RequestingClientID
        {
            get
            {
                return _requestingClientID;
            }
            set
            {
                _requestingClientID = value;
            }
        }

        #endregion

        #region Public Methods

        public void Initialize(Guid clientID, SampleClientSyncProvider clientProvider, ServerSyncProvider serverProvider)
        {
            try
            {
                this.RequestingClientID = clientID;
                this.RemoteProvider = serverProvider;
                this.LocalProvider = clientProvider;

                // We wire to the client here, but it can just easily done in the custom [SampleClientSyncProvider]
                clientProvider.ApplyingChanges += new EventHandler<ApplyingChangesEventArgs>(clientProvider_ApplyingChanges);
                clientProvider.ApplyChangeFailed += new EventHandler<ApplyChangeFailedEventArgs>(clientProvider_ApplyChangeFailed);

                _initialized = true;
            }
            catch (Exception ex)
            {
                this.TrackError("Initialize", ex);
            }
        }
        public List<ConflictLog> DoSynchronize()
        {
            if (!_initialized) { throw new InvalidProgramException("You must call Initialize() before calling DoSynchronize()"); }

            try
            {
                try
                {
                    bool authorized = false;
                    SampleServerSyncProvider serverSync = this.RemoteProvider as SampleServerSyncProvider;
                    if (serverSync != null)
                    {
                        authorized = serverSync.Authorize(RequestingClientID);
                    }
                    if (!authorized)
                    {
                        throw new Exception("Server did not authorize the synchronization");
                    }
                }
                catch (Exception ex)
                {
                    throw new Exception("Unable to connect to server: " + ex.Message, ex);
                }

                //TODO: CLIENT VERIFY connection
                _conflicts = new List<ConflictLog>();
                _syncPolicy = new Dictionary<string, ApplyAction>(StringComparer.OrdinalIgnoreCase);
                //TODO: Get [_syncPolicy] from configuration or preferences, will use default otherwise, see [clientProvider_ApplyChangeFailed]

                // Good spot to log the start
                SyncStatistics stats = this.Synchronize();
                System.Diagnostics.Debug.WriteLine("====================================================");
                System.Diagnostics.Debug.WriteLine("Client: " + this.RequestingClientID.ToString());
                System.Diagnostics.Debug.WriteLine("--------------------");
                System.Diagnostics.Debug.WriteLine("Total download: " + stats.TotalChangesDownloaded);
                System.Diagnostics.Debug.WriteLine("Total download Applied: " + stats.DownloadChangesApplied);
                System.Diagnostics.Debug.WriteLine("Total upload: " + stats.TotalChangesUploaded);
                System.Diagnostics.Debug.WriteLine("Total upload Applied: " + stats.UploadChangesApplied);
                System.Diagnostics.Debug.WriteLine("====================================================");

                // Good spot to log the results
                return _conflicts;
            }
            catch
            {
                // recommend logging
                throw;
            }
        }

        #endregion

        #region Private Methods

        private ApplyAction GetPolicyAction(ConflictType conflictType, ApplyAction defaultAction)
        {
            ApplyAction result = defaultAction;
            if (_syncPolicy.ContainsKey(conflictType.ToString()))
            {
                result = _syncPolicy[conflictType.ToString()];
            }
            return result;
        }

        /// <summary>
        /// Only a placeholder. Should be replaced with custom error mechanisms
        /// </summary>
        private void TrackError(string source, Exception ex)
        {
            //TODO: some smarter error tracking
            MessageBox.Show(ex.Message, source, MessageBoxButtons.OK, MessageBoxIcon.Error);
        }

        #endregion

        #region Event Handlers

        private void clientProvider_ApplyingChanges(object sender, ApplyingChangesEventArgs e)
        {
            try
            {
                // This passes around some conflicts so the server and client can exchange tracking information
                if (e.Session.SyncParameters.Contains("Conflicts"))
                {
                    _conflicts.AddRange((e.Session.SyncParameters["Conflicts"].Value as List<ConflictLog>));
                }
                // likely a good spot to log conflicts
            }
            catch (Exception ex)
            {
                this.TrackError("clientProvider_ApplyingChanges", ex);
            }
        }
        private void clientProvider_ApplyChangeFailed(object sender, ApplyChangeFailedEventArgs e)
        {
            try
            {
                //TODO: This is where you could make a UI for users to make decisions about conflicts.
                //      Be careful however, the related tables will be locked until the sync is completed. 
                //     [so other people won't have access to the tables during the decision]
                ConflictLog conflict = null;
                switch (e.Conflict.ConflictType)
                {
                    case ConflictType.ClientDeleteServerUpdate:
                        // Server Wins (From Policy) Notify User, shouldn't happen based on delete architecture
                        e.Action = GetPolicyAction(e.Conflict.ConflictType, ApplyAction.RetryWithForceWrite);
                        conflict = new ConflictLog(string.Format("A local item was restored from the server. (A recent delete may have been undone). [{0}].  Server chose to: {1}", e.TableMetadata.TableName, e.Action));
                        conflict.Detail.Right = SyncUtility.ConvertToXml(e.Conflict.ServerChange);
                        break;
                    case ConflictType.ClientInsertServerInsert:
                        // Shouldn't happen if using distributed keys, Server wins, notify user.
                        e.Action = GetPolicyAction(e.Conflict.ConflictType, ApplyAction.RetryWithForceWrite);
                        conflict = new ConflictLog(string.Format("Duplicate Primary Keys were detected, The local versions were likely overwritten. Please notify your administrator. [{0}].  Server chose to: {1}", e.TableMetadata.TableName, e.Action));
                        break;
                    case ConflictType.ClientUpdateServerDelete:
                        // Server Wins (From Policy) Notify User, shouldn't happen based on delete architecture
                        e.Action = GetPolicyAction(e.Conflict.ConflictType, ApplyAction.RetryWithForceWrite);
                        conflict = new ConflictLog(string.Format("A local item was deleted by the server. (A recent update may have been lost). [{0}].  Server chose to: {1}", e.TableMetadata.TableName, e.Action));
                        conflict.Detail.Left = SyncUtility.ConvertToXml(e.Conflict.ClientChange);
                        break;
                    case ConflictType.ClientUpdateServerUpdate:
                        // Server Wins (From Policy) Notify User
                        e.Action = GetPolicyAction(e.Conflict.ConflictType, ApplyAction.RetryWithForceWrite);
                        conflict = new ConflictLog(string.Format("A local updated item was updated by the server. (Your update may be lost). [{0}].  Server chose to: {1}", e.TableMetadata.TableName, e.Action));
                        conflict.Detail.Left = SyncUtility.ConvertToXml(e.Conflict.ClientChange);
                        conflict.Detail.Right = SyncUtility.ConvertToXml(e.Conflict.ServerChange);
                        break;
                    case ConflictType.ErrorsOccurred:
                        // Notify User, Cancel
                        conflict = new ConflictLog(string.Format("An unhandled error occured, sync may stop: {0}", e.Conflict.ErrorMessage));
                        break;
                    case ConflictType.Unknown:
                    default:
                        // Notify User, Cancel
                        conflict = new ConflictLog("An unknown conflict occured, sync may stop.");
                        break;
                }
                _conflicts.Add(conflict);
            }
            catch (Exception ex)
            {
                this.TrackError("ApplyChangeFailed", ex);
            }
        }

        #endregion

    }
}
