﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Data;
using System.IO;
using System.Xml;
using System.Data.SqlClient;
using Microsoft.Synchronization.Data;
using Microsoft.Synchronization.Data.SqlExpress;
using Microsoft.Synchronization.Data.Server;
using SQLChangeTrackingSync.Sample.Entities;
using SQLChangeTrackingSync.Sample.Providers;

namespace SQLChangeTrackingSync.Sample
{
    /// <summary>
    /// Utility class to quickly create agents and providers for demonstration purposes.
    /// </summary>
    public static class SyncUtility
    {
        // here for ease only, only used in this class
        public static string ANCHOR_SCHEMA = "dbo"; //NOTE: I typically use a seperate schema for sync tables
        public static string ANCHOR_TABLE = "Anchor";

        public static string ConvertToXml(DataTable dataTable)
        {
            StringBuilder result = new StringBuilder();
            if ((dataTable != null) && (dataTable.Rows.Count > 0))
            {
                using (TextWriter tw = new StringWriter(result))
                {
                    using (XmlTextWriter xw = new XmlTextWriter(tw))
                    {
                        xw.Formatting = Formatting.Indented;
                        dataTable.WriteXml(xw, XmlWriteMode.IgnoreSchema, false);
                    }
                }
            }
            return result.ToString();
        }

        public static SampleClientSyncProvider CreateClientProvider(Guid clientID, string connectionString, List<SillyConfig> tableConfig)
        {
            SqlConnection sqlConnection = new SqlConnection(connectionString);
            SampleClientSyncProvider result = new SampleClientSyncProvider(ANCHOR_SCHEMA, ANCHOR_TABLE);
            result.ClientId = clientID;
            result.Connection = sqlConnection;
            result.SqlChangeTracking = true; // IMPORTANT!

            foreach (SillyConfig item in tableConfig)
            {
                result.SyncAdapters.Add(SyncUtility.GenerateAdapter(sqlConnection, item.schema, item.table, true));
            }
            return result;
        }
        public static SampleServerSyncProvider CreateServerProvider(Guid requestingClientID, string connectionString, List<SillyConfig> tableConfig)
        {
            SqlConnection sqlConnection = new SqlConnection(connectionString);
            SampleServerSyncProvider result = new SampleServerSyncProvider();
            result.RequestingClientID = requestingClientID;
            result.Connection = sqlConnection;

            foreach (SillyConfig item in tableConfig)
            {
                result.SyncAdapters.Add(SyncUtility.GenerateAdapter(sqlConnection, item.schema, item.table, false));
            }

            return result;
        }

        public static SampleSyncAgent CreateSyncAgent(Guid clientID, Guid serverID, string clientConnectionString, string serverConnectionString, List<SillyConfig> tableConfig)
        {
            return CreateSyncAgent(clientID, serverID, clientConnectionString, CreateClientProvider(clientID, clientConnectionString, tableConfig), CreateServerProvider(clientID, serverConnectionString, tableConfig), tableConfig);
        }
        public static SampleSyncAgent CreateSyncAgent(Guid clientID, Guid serverID, string clientConnectionString, SampleClientSyncProvider clientProvider, ServerSyncProvider serverProvider, List<SillyConfig> tableConfig)
        {
            SampleSyncAgent result = new SampleSyncAgent();
            result.Configuration.SyncParameters.Add(SqlExpressClientSyncProvider.PARAMETER_SERVER_ID, serverID);
            result.Configuration.SyncParameters.Add(SqlExpressClientSyncProvider.PARAMETER_ACTUAL_CLIENT_ID, clientID.ToByteArray());

            foreach (SillyConfig item in tableConfig)
            {
                SyncTable syncTable = new SyncTable(string.Format("{0}.{1}", item.schema, item.table));
                syncTable.CreationOption = item.creationOption;
                syncTable.SyncDirection = item.direction;

                result.Configuration.SyncTables.Add(syncTable);
            }

            result.Initialize(clientID, clientProvider, serverProvider);
            return result;
        }
        public static SyncAdapter GenerateAdapter(SqlConnection connection, string tableSchema, string table, bool isClient)
        {
            SqlSyncAdapterBuilder adapterBuilder = new SqlSyncAdapterBuilder(connection);
            adapterBuilder.ChangeTrackingType = ChangeTrackingType.SqlServerChangeTracking;
            adapterBuilder.SyncDirection = SyncDirection.Bidirectional;

            adapterBuilder.TableName = string.Format("{0}.{1}", tableSchema, table);
            SyncAdapter result = adapterBuilder.ToSyncAdapter();
            result.TableName = string.Format("{0}.{1}", tableSchema, table);
            if (isClient)
            {
                result = ApplyServerCollisionFix(result, ANCHOR_SCHEMA, ANCHOR_TABLE);
                result = ApplyAdapterClientIDFix(result);
            }
            return result;
        }

        /// <summary>
        /// String Replacement method to apply fixes to Adapters Generated from the SqlSyncAdapterBuilder. Do not use this
        /// if you manually build your commands.
        /// </summary>
        public static SyncAdapter ApplyServerCollisionFix(SyncAdapter syncAdapter, string anchorSchema, string anchorTable)
        {
            string insertInjection = string.Format("WHERE CT.SYS_CHANGE_CREATION_VERSION <> (SELECT ISNULL([anc_server_applied],-1) " +
                "FROM [{0}].[{1}] WHERE [anc_table] = '{2}') AND ", anchorSchema, anchorTable, syncAdapter.TableName);
            syncAdapter.SelectIncrementalInsertsCommand.CommandText = syncAdapter.SelectIncrementalInsertsCommand.CommandText.Replace("WHERE ", insertInjection);

            string updateInjection = string.Format("DECLARE @ServerAppliedAnchor as BIGINT \r\n" +
                    "SELECT @ServerAppliedAnchor = ISNULL([anc_server_applied],[anc_sent]) " +
                    "FROM [{0}].[{1}] WHERE [anc_table] = '{2}' \r\n", anchorSchema, anchorTable, syncAdapter.TableName);
            syncAdapter.SelectIncrementalUpdatesCommand.CommandText = updateInjection + syncAdapter.SelectIncrementalUpdatesCommand.CommandText.Replace(", @sync_last_received_anchor", ", @ServerAppliedAnchor");

            return syncAdapter;
        }

        /// <summary>
        /// String replacement method to apply SQL Generation Fixes to SqlSyncAdapterBuilder when used as a client. Do not use this
        /// if you manually build your commands.
        /// </summary>
        public static SyncAdapter ApplyAdapterClientIDFix(SyncAdapter syncAdapter)
        {
            (syncAdapter.InsertCommand as SqlCommand).Parameters.Add("@" + SqlExpressClientSyncProvider.PARAMETER_ACTUAL_CLIENT_ID, SqlDbType.VarBinary, 16);
            (syncAdapter.UpdateCommand as SqlCommand).Parameters.Add("@" + SqlExpressClientSyncProvider.PARAMETER_ACTUAL_CLIENT_ID, SqlDbType.VarBinary, 16);
            (syncAdapter.DeleteCommand as SqlCommand).Parameters.Add("@" + SqlExpressClientSyncProvider.PARAMETER_ACTUAL_CLIENT_ID, SqlDbType.VarBinary, 16);
            syncAdapter.InsertCommand.CommandText = syncAdapter.InsertCommand.CommandText.Replace(@"@sync_client_id_binary", "@" + SqlExpressClientSyncProvider.PARAMETER_ACTUAL_CLIENT_ID);
            syncAdapter.UpdateCommand.CommandText = syncAdapter.UpdateCommand.CommandText.Replace(@"@sync_client_id_binary", "@" + SqlExpressClientSyncProvider.PARAMETER_ACTUAL_CLIENT_ID);
            syncAdapter.DeleteCommand.CommandText = syncAdapter.DeleteCommand.CommandText.Replace(@"@sync_client_id_binary", "@" + SqlExpressClientSyncProvider.PARAMETER_ACTUAL_CLIENT_ID);
            return syncAdapter;
        }
       
    }
}
